import datetime
import os
from tkinter import Pack
from typing import List
from git import Repo, Commit, Remote
import requests
import sys
import json

LAST_RELEASE_VERSION = "LAST_RELEASE_VERSION"
LAST_RELEASE_HASH = "LAST_RELEASE_HASH"

def set_value_from_api(api_token, project_id, key, value):
    url = "https://gitlab.evosoft.xyz/api/v4/projects/{}/variables/{}".format(project_id, key)

    payload='key={}&value={}&protected=false&masked=false'.format(key, value)

    headers = {
        'PRIVATE-TOKEN': api_token,
        'Content-Type': 'application/x-www-form-urlencoded'
    }

    response = requests.put(url, headers=headers, data=payload)

    if response.status_code != 200:
        raise Exception("Не удалось обновить ключ {}".format(key))
    else:
        print("Обновлен ключ {} со значением {}.".format(key, value))

def create_value_from_api(api_token, project_id, key, value):
    url = "https://gitlab.evosoft.xyz/api/v4/projects/{}/variables/".format(project_id)

    payload='key={}&value={}&protected=false&masked=false'.format(key, value)

    headers = {
        'PRIVATE-TOKEN': api_token,
        'Content-Type': 'application/x-www-form-urlencoded'
    }

    response = requests.post(url, headers=headers, data=payload)

    if response.status_code != 201:
        raise Exception("Не удалось создать ключ {}".format(key))
    else:
        print("Создан ключ {} со значением {}.".format(key, value))

def get_value_from_api(api_token, project_id, key, default_value):
    url = "https://gitlab.evosoft.xyz/api/v4/projects/{}/variables/{}".format(project_id, key)

    # glpat-6uYCYZQS9jNnbrhmn4nn
    headers = {
        'PRIVATE-TOKEN': api_token
    }

    response = requests.get(url, headers=headers)

    if response.status_code == 200:
        value = response.json()['value']
        print("Получен ключ {} со значением {}.".format(key, value))
        return value
    if response.status_code == 404:
        create_value_from_api(api_token, project_id, key, default_value)
        print("Получен ключ {} со значением {}.".format(key, default_value))
        return default_value
    else:
        raise Exception("Не удалось получить данные о ключе {}".format(key))

def get_split_version(version):
    POST = 0
    PATCH = 0
    MINOR = 0
    MAJOR = 0

    version = version.strip().split('.')

    try:
        MAJOR = int(version[0])
    except:
        MAJOR = 0

    try:
        MINOR = int(version[1])
    except:
        MINOR = 1

    try:
        PATCH = int(version[2])
    except:
        PATCH = 0

    try:
        POST = int(version[3])
    except:
        POST = 0

    return (MAJOR, MINOR, PATCH, POST)

def get_last_release_from_api(api_token, project_id):
    version = get_value_from_api(api_token, project_id, LAST_RELEASE_VERSION, "0.1.0")
    hexsha = get_value_from_api(api_token, project_id, LAST_RELEASE_HASH, ".")

    return ( get_split_version(version), hexsha )

def get_all_commits() -> List[Commit]:
    repo = Repo('.')
    commits = list(repo.iter_commits('--all'))

    items = [ commit for commit in commits if (commit.message.lower().count("merge branch") <= 0) ][::-1]
    
    print("Commits count: {0}".format(len(items)))

    return items

def get_new_version(last_version, last_hexsha):
    MAJOR = last_version[0]
    MINOR = last_version[1]
    PATCH = last_version[2]
    POST = last_version[3]

    version = ''
    hexsha = last_hexsha

    for commit in get_all_commits():
        name = commit.message.replace("\n", "")
        hexsha = commit.hexsha
        email = commit.author.email
        committed_datetime = commit.committed_datetime

        POST = POST + 1

        if (hexsha == last_hexsha):
            MAJOR = last_version[0]
            MINOR = last_version[1]
            PATCH = last_version[2]
            POST = last_version[3] + 1
        else:
            if (name.lower().count("fix") > 0):
                POST = 0
                PATCH = PATCH + 1

            if (name.lower().count("feat") > 0):
                POST = 0
                PATCH = 0
                MINOR = MINOR + 1

            if (name.lower().count("breaking change") > 0):
                POST = 0
                PATCH = 0
                MINOR = 0
                MAJOR = MAJOR + 1
        
        version = "{0}.{1}".format(MAJOR, MINOR)

        if (PATCH != 0):
            version = "{0}.{1}".format(version, PATCH)

        if (POST != 0):
            version = "{0}.{1}".format(version, POST)

        print("{}: {}: ({}:{}) {}".format(version, hexsha, email, committed_datetime, name))
    return version

def get_new_commits(last_hexsha):
    commits = list()

    for commit in get_all_commits():
        name = commit.message.replace("\n", "")
        hexsha = commit.hexsha
        email = commit.author.email
        committed_datetime = commit.committed_datetime

        if (hexsha == last_hexsha):
            commits.clear()

        commits.append({
            "id": hexsha,
            "message": name,
            "author_email": email,
            "timestamp": str(committed_datetime)
        })

    return commits

def get_new_hash():
    commits = list(get_all_commits())

    if len(commits) <= 0:
        return "."
    else:
        return commits[::-1][0].hexsha

def create_new_release_to_api(api_token, project_id, new_version, new_hash):
    set_value_from_api(api_token, project_id, LAST_RELEASE_VERSION, new_version)
    set_value_from_api(api_token, project_id, LAST_RELEASE_HASH, new_hash)

def create_new_tag(api_token, project_id, branch_name, version, commits):
    message = ""
    for commit in commits:
        message = "{} [{}] {}\n{}".format(commit['timestamp'], commit['author_email'], commit['message'], message)

    url = "https://gitlab.evosoft.xyz/api/v4/projects/{}/repository/tags?tag_name={}&ref={}&message={}".format(project_id, version, branch_name, message)

    headers = {
        'PRIVATE-TOKEN': api_token,
        'Content-Type': 'application/x-www-form-urlencoded'
    }

    requests.post(url, headers=headers)

def create_new_release_gitlab(api_token, project_id, version):
    url = "https://gitlab.evosoft.xyz/api/v4/projects/{}/releases".format(project_id)

    data = {
        'tag_name': version
    }
    
    headers = {
        'PRIVATE-TOKEN': api_token,
        'Content-Type': 'application/x-www-form-urlencoded'
    }

    requests.post(url, headers=headers, data=data)

if __name__ == "__main__":
    print(f"Arguments count: {sys.argv}")

    for key, value in os.environ.items():
        print("'{}': '{}'".format(key, value))

    # $CI_PROJECT_TITLE
    project_name = os.environ['CI_PROJECT_TITLE']
    # $API_ACCESS_TOKEN
    api_token = os.environ['API_ACCESS_TOKEN']
    # $CI_PROJECT_ID
    project_id = os.environ['CI_PROJECT_ID']
    # $CI_COMMIT_BRANCH
    branch_name = os.environ['CI_COMMIT_BRANCH']

    try:
        last_version = os.environ['LAST_RELEASE_VERSION']
    except:
        last_version = ""

    try:
        last_hexsha = os.environ['LAST_RELEASE_HASH']
    except:
        last_hexsha = ""

    version = get_new_version(last_version, last_hexsha)
    commits = get_new_commits(last_hexsha)

    create_new_release_to_api(api_token, project_id, version, get_new_hash())
    create_new_tag(api_token, project_id, branch_name, version, commits)
    create_new_release_gitlab(api_token, project_id, version)

    with open('.next-version', 'w') as version_file:
        version_file.write(version)

    print("next-version: {0}".format(version))
    print("next-commits: {0}".format(commits))
